/* arc4gen - Generate or verify an ARC4 stream */
/* Written 2004-05-14 by David A. Madore <david.madore@ens.fr> */
/* Public Domain */

/* This program generates or checks an ARC4 stream (cryptographically
 * strong random number generator). */

/* Compile with something like
    gcc -o arc4gen arc4gen.c -O6 -Wall
 * If you have GNU getopt around, you may wish to add
 * -DWITH_LONG_OPTIONS=1 to enable GNU-style long options. */

/* Syntax for use is:
    arc4gen [-c] [-k key | -K hexkey] [-s size] [filename]
 * Long option alternatives are: --check for -c, --key for -k,
 * --hexkey for -K and --size for -s. */

/* arc4gen will generate an ARC4 stream and write it out to filename,
 * except when -c is passed as option, in which case it checks that
 * filename does contain an identical copy of the ARC4 stream.  If
 * filename is omitted, the standard output or input is understood
 * instead (input when -c is passed, output otherwise).  The -s option
 * can be used to specify a bound on the size (the suffixes 'k' and
 * 'M' are understood to refer to 1024 and 1048576 bytes); if this
 * option is not given, then the stream is endless upon writing (so it
 * is probably not a good idea to use this with a filename, unless you
 * wish to fill your hard drive) and upon checking proceeds to the end
 * of file. */

/* The -k or -K options can be used to specify the coding key, which
 * otherwise is null.  If -k is specified, the following argument is
 * understood to be the key verbatim; if -K is specified, it is
 * understood to be the key in hexadecimal.  For example, it is
 * equivalent to give -k foobar or -K 666f6f626172 as argument.  In
 * either case, the key is repeated as necessary to fill the 256 bytes
 * of the ARC4 key bins. */

/* Return code is: 0 in case of success (either because the stream has
 * been checked sucessfully or because the required number of bytes
 * has been written), 1 in case of failed check, 2 in case of system
 * error and 3 in case of incorrect usage. */

/* Here are a few possible uses of this program: */
/* Basic sanity check:
    arc4gen -s 16M | md5sum
 * should return f949bf88fd4a13f090592d1b18cbe098 (assuming you have
 * the md5sum program available, of course...). */
/* To check that your CPU is consistent, run:
    arc4gen | arc4gen -c
 * (or possibly add a -k something argument to both arc4gens) and let
 * it run for hours. */
/* To check memory, you might try running:
    arc4gen -s 128M somefile
 * and then
    arc4gen -c somefile
 * repeatedly.  With luck, the 128Mb file (adjust size according to
 * your amount of RAM: tune so that the disk is used as little as
 * possible) will remain in cache all the time, and most of the RAM
 * will be checked. */
/* You might also check disks and network in similar ways, but it is
 * probably useless since they have adequate error detecting codes.  A
 * possible filesystem check might be:
    for i in `seq 1000` ; do arc4gen -s 64k -k "$i" "$i.chk" ; done
    for f in *.chk ; do arc4gen -c -k "${f%.chk}" "$f" ; done
 */

/* Important note: the ARC4 generator is known to be statistically
 * biased (toward producing the same byte twice in a row: this happens
 * with a probability of 2^-8 + 2^-24 approximately, rather than 2^-8
 * if consecutive bytes were independent and uniformly distributed).
 * Therefore, do not use it for scientific works requiring high
 * quality random streams.  Also, do not use it for generating
 * one-time pads and similar sensitive data. */

/* Short description of the ARC4 algorithm: the state is described by
 * two 8-bit quantities, i and j, and a permutation of 0..255, the
 * "S-boxes", S[0] through S[255].  To produce one byte: (1) increment
 * i mod 256, (2) increase j by S[i] mod 256, (3) swap S[i] and S[j],
 * and (4) output S[S[i]+S[j] mod 256].  To initialize the generator
 * from the 256 key bytes K[0] through K[255]: initialize i=0, j=0,
 * S[0]=0, S[1]=1 through S[255]=255; then run i from 0 through 255
 * and, for each value, (1) increase j by S[i]+K[i] and (2) swap S[i]
 * and S[j]; then reset i and j back to 0. */

/* RC4 is claimed to be a trade secret.  Ha, ha, ha.  Very funny.
 * ARC4 is the non-secret version which by a weird coincidence just so
 * happens to appear to do exactly the same thing as RC4, but you're
 * not allowed to say that, so you're not allowed to say that this
 * program actually computes RC4.  You're also not allowed to say that
 * this intellectual property stuff is bullshit, I suppose. */

#ifndef WITH_LONG_OPTIONS
#define WITH_LONG_OPTIONS 0
#endif

#ifndef DEBUG
#define DEBUG 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#if WITH_LONG_OPTIONS
#include <getopt.h>
#endif

/* For handling of options */
#if WITH_LONG_OPTIONS
const struct option longopts[] = {
  { "check", no_argument, 0, 'c' },
  { "key", required_argument, 0, 'k' },
  { "hexkey", required_argument, 0, 'K' },
  { "size", required_argument, 0, 's' },
};
#endif
const char *shortopts = "ck:K:s:";

/* ARC4 generator state */
unsigned char arc4key[256];
unsigned char arc4sboxes[256];
unsigned char arc4i, arc4j;

unsigned char
arc4byte (void)
     /* Produce one byte of the arc4 stream. */
{
  unsigned char t;

  arc4i = (arc4i+1)&0xff;
  arc4j = (arc4j+arc4sboxes[arc4i])&0xff;
  t = arc4sboxes[arc4i];
  arc4sboxes[arc4i] = arc4sboxes[arc4j];
  arc4sboxes[arc4j] = t;
  return arc4sboxes[(arc4sboxes[arc4i]+arc4sboxes[arc4j])&0xff];
}

void
arc4keyinit (const char *key)
     /* Set the ARC4 generator to the given key (max 256 bytes).
      * Remember to call arc4init() after this! */
{
  int i, i0;

  i0 = 0;
  for ( i=0 ; i<256 ; i++ )
    {
      if ( ! key[i0] )
	{
	  if ( i0 == 0 )
	    {
	      /* PUKE! */
	      arc4key[i] = 0;
	      continue;
	    }
	  i0 = 0;
	}
      arc4key[i] = key[i0++];
    }
}

unsigned char
hexvalue (char dig)
{
  if ( dig >= '0' && dig <= '9' )
    return dig-'0';
  else if ( dig >= 'a' && dig <= 'f' )
    return (dig-'a')+10;
  else if ( dig >= 'A' && dig <= 'F' )
    return (dig-'A')+10;
  else
    {
      fprintf (stderr, "Invalid hexadecimal character!\n");
      exit (3);
    }
}

void
arc4hexkeyinit (const char *hexkey)
     /* Set the ARC4 generator to the given key (max 256 bytes), which
      * is given in hexadecimal.  Remember to call arc4init() after
      * this! */
{
  int i, i0;

  i0 = 0;
  for ( i=0 ; i<256 ; i++ )
    {
      if ( ! hexkey[i0] )
	i0 = 0;
      if ( ! ( hexkey[i0] && hexkey[i0+1] ) )
	{
	  fprintf (stderr, "Hexadecimal string ends prematurely!\n");
	  exit (3);
	}
      arc4key[i] = (hexvalue (hexkey[i0]) << 4)
	+ hexvalue (hexkey[i0+1]);
      i0 += 2;
    }
}

void
arc4init (void)
     /* Initialize the ARC4 generator. */
{
  int i, j;
  unsigned char t;

#if DEBUG
  fprintf (stderr, "DEBUG: ARC4 key bytes are as follows:\n");
  for ( i=0 ; i<256 ; i++ )
    {
      fprintf (stderr, "%02x", arc4key[i]);
      if ( i%32 == 31 )
	fprintf (stderr, "\n");
    }
#endif
  for ( i=0 ; i<256 ; i++ )
    arc4sboxes[i] = i;
  arc4i = 0;
  arc4j = 0;
  j = 0;
  for ( i=0 ; i<256 ; i++ )
    {
      j = (j+arc4sboxes[i]+arc4key[i])&0xff;
      t = arc4sboxes[i];
      arc4sboxes[i] = arc4sboxes[j];
      arc4sboxes[j] = t;
    }
#if DEBUG
  fprintf (stderr, "DEBUG: ARC4 S-boxes are as follows:\n");
  for ( i=0 ; i<256 ; i++ )
    {
      fprintf (stderr, "%02x", arc4sboxes[i]);
      if ( i%32 == 31 )
	fprintf (stderr, "\n");
    }
#endif
}

void
usage (const char progname[])
{
  fprintf (stderr,
	   "Usage is: %s [-c] [-k key | -K hexkey] [-s size] [filename]\n",
	   progname);
  exit (3);
}

int
main (int argc, char *argv[])
{
  char *tail;
  char check;
  FILE *f;
  unsigned char buf[BUFSIZ];
  size_t size;
  size_t pos;
  size_t usesz;
  size_t sz;
  size_t i;

  check = 0;
  size = 0;
  arc4keyinit ("");
  while ( 1 )
    {
      int c;


#if WITH_LONG_OPTIONS
      c = getopt_long (argc, argv, shortopts, longopts, NULL);
#else
      c = getopt (argc, argv, shortopts);
#endif
      if ( c == -1 )
	break;
      switch ( c )
	{
	case 'c':
	  check = 1;
	  break;
	case 'k':
	  arc4keyinit (optarg);
	  break;
	case 'K':
	  arc4hexkeyinit (optarg);
	  break;
	case 's':
	  size = strtol (optarg, &tail, 0);
	  if ( strcmp (tail, "") == 0 )
	    ;
	  else if ( strcmp (tail, "k") == 0 )
	    size *= 1024;
	  else if ( strcmp (tail, "M") == 0 )
	    size *= 1024*1024;
	  else
	    {
	      fprintf (stderr, "Argument to -s not understood\n");
	      exit (3);
	    }
	  break;
	default:
	  usage (argv[0]);
	}
    }
  if ( optind+1 < argc )
    usage (argv[0]);
  pos = 0;
  arc4init ();
  if ( check )
    {
      if ( optind < argc )
	{
	  f = fopen (argv[optind], "rb");
	  if ( ! f )
	    {
	      fprintf (stderr, "Error trying to open %s for reading\n", argv[optind]);
	      perror ("fopen");
	      exit (2);
	    }
	}
      else
	f = stdin;
      while ( 1 )
	{
	  if ( ferror (f) )
	    {
	      fprintf (stderr, "Error while reading\n");
	      perror ("ferror");
	      exit (2);
	    }
	  if ( feof (f) )
	    exit (0);
	  if ( size )
	    {
	      usesz = size-pos;
	      if ( usesz > BUFSIZ )
		usesz = BUFSIZ;
	    }
	  else
	    usesz = BUFSIZ;
	  if ( ! usesz )
	    exit (0);
	  sz = fread (buf, 1, usesz, f);
	  for ( i=0 ; i<sz ; i++ )
	    if ( buf[i] != arc4byte () )
	      {
		fprintf (stderr, "Check failed at byte %llu (%0#llx)\n",
			 (long long unsigned)(pos+i), (long long unsigned)(pos+i));
		exit (1);
	      }
	  pos += sz;
	}
    }
  else
    {
      if ( optind < argc )
	{
	  f = fopen (argv[optind], "wb");
	  if ( ! f )
	    {
	      fprintf (stderr, "Error trying to open %s for writing\n", argv[optind]);
	      perror ("fopen");
	      exit (2);
	    }
	}
      else
	f = stdout;
      while ( 1 )
	{
	  if ( ferror (f) )
	    {
	      fprintf (stderr, "Error while writing\n");
	      perror ("ferror");
	      exit (2);
	    }
	  if ( size )
	    {
	      usesz = size-pos;
	      if ( usesz > BUFSIZ )
		usesz = BUFSIZ;
	    }
	  else
	    usesz = BUFSIZ;
	  if ( ! usesz )
	    exit (0);
	  for ( i=0 ; i<usesz ; i++ )
	    buf[i] = arc4byte ();
	  sz = fwrite (buf, 1, usesz, f);
	  if ( sz < usesz )
	    fprintf (stderr, "Short write!\n");
	  pos += sz;
	}
    }
}
